#lang racket
(require "test.rkt")

;; An any+stream is (cons any list-of-string)
;; A list+stream is (cons list list-of-string)

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda ()'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; parse : sym table list-of-string -> any+stream-or-false
(define (parse what table stream)
  (define defn (assq what table))
  (if defn
      (parse-alts (rest defn) table stream)
      (error 'parse "no definition for ~a" what)))

(test (parse 'main table '("apple")) (cons 'apple '()))
(test (parse 'fruit-salad table '("orange" "orange")) 
      (cons '(orange orange) '()))
(test (parse 'main table '("orange" "orange" "!")) 
      (cons '((orange orange) !) '()))
#;(test (parse 'main table '("orange" "orange" "orange" "!"))  
        (cons '((orange orange orange) !) '()))
(test (parse 'main table '("radish")) #f)

;; parse-alts : list-of-case table list-of-string -> any+stream-or-false
(define (parse-alts alts table stream)
  (if (null? alts)
      false
      (or (parse-case (first alts) table stream)
          (parse-alts (rest alts) table stream))))

(let ([alts (list
             `[("apple") ,(lambda () 'apple)]
             `[("grape") ,(lambda () 'grape)])])
  (test (parse-alts alts table '("apple")) (cons 'apple '()))
  (test (parse-alts alts table '("grape")) (cons 'grape '()))
  (test (parse-alts alts table '("grape" "fish")) (cons 'grape '("fish")))
  (test (parse-alts alts table '("radish")) #f))

;; parse-case : case table list-of-string -> any+stream-or-false
(define (parse-case case table stream)
  (define seq (first case))
  (define handler (second case))
  (define parsed+stream (parse-sequence seq table stream))
  (and parsed+stream
       (cons (apply handler (car parsed+stream))
             (cdr parsed+stream))))

(let ()
  (define case `[("apple") ,(lambda () 'apple)])
  (test (parse-case case table '("apple")) (cons 'apple '()))
  (test (parse-case case table '("grape")) #f))

;; parse-sequence : sequence table list-of-string -> list-or-false
(define (parse-sequence seq table stream)
  (cond
    [(empty? seq) (cons empty stream)]
    [else (define first-parsed+stream
            (parse-item (first seq) table stream))
          (and first-parsed+stream
               (let ()
                 (define rest-parsed+stream
                   (parse-sequence (rest seq) table 
                                   (cdr first-parsed+stream)))
                 (and rest-parsed+stream
                      (cons (append (car first-parsed+stream)
                                    (car rest-parsed+stream))
                            (cdr rest-parsed+stream)))))]))

(let ()
  (define seq `("apple" "grape"))
  (test (parse-sequence '("apple") table '("apple")) (cons empty '()))
  (test (parse-sequence seq table '("apple" "grape")) (cons empty '()))
  (test (parse-sequence seq table '("apple" "radish")) #f))
(let ()
  (define seq '(fruit fruit))
  (test (parse-sequence seq table '("orange" "grape"))
        (cons '(orange grape) '()))
  (test (parse-sequence seq table '("grape" "orange"))
        (cons '(grape orange) '())))

;; parse-item : item table list-of-string -> list+stream-or-false
(define (parse-item item table stream)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (cons empty (cdr stream))
         false)]
    [else
     (define x+stream (parse item table stream))
     (and x+stream 
          (cons (list (car x+stream))
                (cdr x+stream)))]))

(test (parse-item '"grape" table '("grape")) (cons empty '()))
(test (parse-item 'fruit table '("grape")) (cons '(grape) '()))
(test (parse-item 'fruit table '("radish")) #f)

(run-tests)
