#lang racket
(require "test.rkt")

;; A tree is either
;;  - string
;;  - (list string tree tree)

;; find-path : tree string -> false-or-list-of-string
(define (find-path t name)
  (cond
    [(string? t) (if (equal? t name)
                     empty
                     #f)]
    [else      
     (let ([left-p (find-path (cadr t) name)])
       (if left-p
           (cons (car t) left-p)
           (let ([right-p (find-path (caddr t) name)])
             (if right-p
                 (cons (car t) right-p)
                 #f))))]))

(test (find-path '("a" ("b" "1" "2") ("c" "3" "4")) "3") '("a" "c"))
(test (find-path '("a" ("b" "1" "2") ("c" "3" "4")) "5") #f)



;; When building up a result from recursive calls, to use a failure
;;  continuation, you also need a success continuation:

;; A success is (list-of-string -> alpha)
;; A fail is (-> alpha)

;; find-path* : tree string success fail -> alpha
(define (find-path* t name success-k fail-k)
  (cond
    [(string? t) (if (equal? t name)
                     (success-k empty)
                     (fail-k))]
    [else
     (find-path* (cadr t) name
                 (lambda (p)
                   (success-k (cons (car t) p)))
                 (lambda ()
                   (find-path* (caddr t) name
                               (lambda (p) (success-k (cons (car t) p)))
                               fail-k)))]))

(define (ok l) l)
(define (nope) #f)

(test (find-path* '("a" ("b" "1" "2") ("c" "3" "4")) "3" ok nope) 
      '("a" "c"))
#|
=>
(find-path* '("b" "1" "2") "3" 
            (lambda (p) (ok (cons "a" p)))
            (lambda ()
              (find-path* '("c" "3" "4") "3" 
                          (lambda (p) (ok (cons "a" p)))
                          nope)))
=>
(find-path* "1" "3"
            (lambda (p)
              ((lambda (p) (ok (cons "a" p)))
               (cons "b" p)))
            (lambda ()
              (find-path* "2" "3"
                          (lambda (p)
                            ((lambda (p) (ok (cons "a" p)))
                             (cons "b" p)))
                          (lambda ()
                            (find-path* '("c" "3" "4") "3" 
                                        (lambda (p) (ok (cons "a" p)))
                                        nope)))))
=> 
(find-path* '("c" "3" "4") "3" 
            (lambda (p) (ok (cons "a" p)))
            nope)
=>
(find-path* "3" "3"
            (lambda (p) ((lambda (p) (ok (cons "a" p)))
                         (cons "c" p)))
            (lambda ()
              (find-path* "4" "3"
                          (lambda (p) ((lambda (p) (ok (cons "a" p)))
                                       (cons "c" p)))
                          nope)))
=>
((lambda (p) ((lambda (p) (ok (cons "a" p)))
              (cons "c" p)))
 null)
|#

(test (find-path* '("a" ("b" "1" "2") ("c" "3" "4")) "5" ok nope) #f)

(run-tests)
